@@ -58,38 +58,66 @@ typedef struct dbinfo {	/* file information for the database */
 
 typedef struct fileinfo {	/* database file information */
   dbinfo	dbinf;
+  char		*path;		/* null-terminated */
+  size_t	pathlen;	/* save on strlen calls,
+				 * since we know it already */
+  unsigned long	ruleset;
+  unsigned	found_ruleset: 1; /* flag indicates whether ruleset
+				   * member has been initialized */
   unsigned	did_sum: 1;	/* flag shows whether sum member
 				 * contains a valid checksum */
 } fileinfo;
 
-inline static void do_checksum(dbinfo *dbinf, const char *fname)
+static void fileinfo_init(fileinfo *inf, char *path, size_t pathlen)
+{
+    inf->ruleset	 = 0;
+    inf->found_ruleset	 = 0;
+    inf->did_sum	 = 0;
+    inf->path		 = path;
+    inf->pathlen	 = pathlen;
+}
+
+static unsigned long fileinfo_ruleset(options *opts, fileinfo *inf)
+{
+    if (! inf->found_ruleset) {
+      inf->ruleset		 = rules_for_path(opts, inf->path);
+      inf->found_ruleset	 = 1;
+    }
+
+    return inf->ruleset;
+}
+
+inline static void do_checksum(options *opts, fileinfo *inf)
 {
     SHA_CTX	context;
     char	buf[BUFSIZ];
     int		n;
-    int		fd	 = open(fname, O_RDONLY);
-#ifdef ELC_RESET_ATIME
+    int		fd	 = open(inf->path, O_RDONLY);
+    dbinfo	*dbinf	 = &inf->dbinf;
     struct utimbuf	utb;	/* for resetting access time */
-#endif
 
     if (fd == -1)
       die(__FUNCTION__, "Error: opening file (%s): %s",
-	  fname, strerror(errno));
+	  inf->path, strerror(errno));
     SHA1_Init(&context);
     while ( (n = read(fd, buf, BUFSIZ)) )
       SHA1_Update(&context, buf, n);
     close(fd);
     SHA1_Final(dbinf->sum, &context);
 
-#ifdef ELC_RESET_ATIME
-    /* reset access time */
-    utb.actime	 = dbinf->stat.st_atime;
-    utb.modtime	 = dbinf->stat.st_mtime;
-    if (utime(fname, &utb) == -1)
-      warn(__FUNCTION__,
-	  "Warning: resetting access time for file (%s): %s",
-	  fname, strerror(errno));
+    /* reset access time on request */
+    if (fileinfo_ruleset(opts, inf) & RULE_RESET_ATIME) {
+#ifdef	DEBUG
+      fprintf(stderr, "debug (%s): resetting atime for file (%s)\n",
+	      __FUNCTION__, inf->path);
 #endif
+      utb.actime	 = dbinf->stat.st_atime;
+      utb.modtime	 = dbinf->stat.st_mtime;
+      if (utime(inf->path, &utb) == -1)
+	warn(__FUNCTION__,
+	     "Warning: resetting access time for file (%s): %s",
+	     inf->path, strerror(errno));
+    }
 }
 
 static void show_diff_xml_long(FILE *out, char T, const char *type,
@@ -162,6 +190,9 @@ static void show_diffs(options *opts,
     void (*show_oct)(FILE *out, char T, const char *type,
 		     unsigned long old, unsigned long new);
 
+    /* for showing the last twelve bits (four octal digits) */
+    const	unsigned perm_mask	 = 07777;
+
     if (opts->output == OUTPUT_XML) {
       show_long	 = show_diff_xml_long;
       show_time	 = show_diff_xml_long;
@@ -177,7 +208,8 @@ static void show_diffs(options *opts,
     if (diffs & RULE_INODE)
       show_long(stdout, 'i', "inode", sa->st_ino, sb->st_ino);
     if (diffs & RULE_PERMS)
-      show_oct(stdout, 'p', "permissions", sa->st_mode, sb->st_mode);
+      show_oct(stdout, 'p', "permissions",
+	       sa->st_mode & perm_mask, sb->st_mode & perm_mask);
     if (diffs & RULE_NLINK)
       show_long(stdout, 'l', "nlinks", sa->st_nlink, sb->st_nlink);
     if (diffs & RULE_UID)
@@ -190,10 +222,8 @@ static void show_diffs(options *opts,
       show_time(stdout, 'a', "access_time", sa->st_atime, sb->st_atime);
     if (diffs & RULE_MTIME)
       show_time(stdout, 'm', "modification_time", sa->st_mtime, sb->st_mtime);
-#if 0
     if (diffs & RULE_CTIME)
       show_time(stdout, 'c', "change_time", sa->st_ctime, sb->st_ctime);
-#endif
 
     if (opts->output == OUTPUT_XML)
       XML_END_PRINT(stdout, "change");
@@ -202,11 +232,12 @@ static void show_diffs(options *opts,
 }
 
 static void report_stat_differences(options *opts,
-				    const char *path, unsigned long flags,
 				    fileinfo *currinf, dbinfo *old)
 {
     struct stat		*sa	 = &old->stat;
     struct stat		*sb	 = &currinf->dbinf.stat;
+    const char *path		 = currinf->path;
+    unsigned long	flags	 = fileinfo_ruleset(opts, currinf);
     unsigned long	diffs	 = 0;
     
     if ((flags & RULE_INODE)
@@ -233,16 +264,16 @@ static void report_stat_differences(options *opts,
     if ((flags & RULE_MTIME)
 	&& (sa->st_mtime != sb->st_mtime))
       diffs	 |= RULE_MTIME;
-#if 0
-    if ((flags & RULE_CTIME)
+    if ((flags & RULE_CTIME) && (! (flags & RULE_RESET_ATIME))
 	&& (sa->st_ctime != sb->st_ctime))
       diffs	 |= RULE_CTIME;
-#endif
 
     if (diffs)
       show_diffs(opts, path, diffs, sa, sb);
 }
 
+/* report that a file has changed from a non-regular file
+ * to a regular file */
 static void report_2regfile(options *opts, const char *path)
 {
     switch (opts->output) {
@@ -261,6 +292,7 @@ static void report_2regfile(options *opts, const char *path)
     }
 }
 
+/* report that the checksum has changed. */
 static void report_sumchange(options *opts, const char *path,
 			     const unsigned char *old, size_t oldsiz,
 			     const unsigned char *new, size_t newsiz)
@@ -289,13 +321,13 @@ static void report_sumchange(options *opts, const char *path,
     }
 }
 
-static void report_differences(options *opts, fileinfo *currinf,
-				      const char *path)
+static void report_differences(options *opts, fileinfo *currinf)
 {
     dbinfo		old;
     struct cdb		*knowndb	 = &opts->knowndb;
     size_t		knownsiz	 = cdb_datalen(knowndb);
-    unsigned long	flags		 = rules_for_path(opts, path);
+    char		*path		 = currinf->path;
+    unsigned long	flags		 = fileinfo_ruleset(opts, currinf);
     
     if (knownsiz != sizeof(old)
 	&& knownsiz != sizeof(old.stat))
@@ -308,7 +340,7 @@ static void report_differences(options *opts, fileinfo *currinf,
 	report_2regfile(opts, path);
       else if (flags & RULE_SUM) {
 	if (! currinf->did_sum) {
-	  do_checksum(&currinf->dbinf, path);
+	  do_checksum(opts, currinf);
 	  currinf->did_sum	 = 1;
 	}
 	if (memcmp(currinf->dbinf.sum, old.sum, sizeof(old.sum)))
@@ -316,11 +348,13 @@ static void report_differences(options *opts, fileinfo *currinf,
 			   currinf->dbinf.sum, sizeof(currinf->dbinf.sum));
       }
     }
-    report_stat_differences(opts, path, flags, currinf, &old);
+    report_stat_differences(opts, currinf, &old);
 }
 
-static void report_newfile(options *opts, const char *path)
+static void report_newfile(options *opts, fileinfo *inf)
 {
+    char	*path	 = inf->path;
+
     switch (opts->output) {
       case OUTPUT_XML:
 	XML_CHANGE_START_PRINT(stdout, "newfile", path);
@@ -336,51 +370,53 @@ static void report_newfile(options *opts, const char *path)
     }
 }    
 
-inline static void do_check(options *opts,
-			    char *path, size_t pathlen, fileinfo *inf)
+inline static void do_check(options *opts, fileinfo *inf)
 {
     struct cdb	*db	 = &opts->knowndb;
     int		err;
 
-    if ( (err = cdb_find(db, path, pathlen)) == -1)
+    if ( (err = cdb_find(db, inf->path, inf->pathlen)) == -1)
       die(__FUNCTION__,
 	  "Error: looking up file (%s) in known database (%s): %s",
-	  path, opts->knowndbname, strerror(errno));
+	  inf->path, opts->knowndbname, strerror(errno));
     else if (!err)
-      report_newfile(opts, path);
+      report_newfile(opts, inf);
     else
-      report_differences(opts, inf, path);
+      report_differences(opts, inf);
 }
 
-inline static void do_update(options *opts,
-			     char *path, size_t pathlen, fileinfo *inf)
+inline static void do_update(options *opts, fileinfo *inf)
 {
     struct cdb_make	*db	 = &opts->currdb;
     dbinfo		*dbinf	 = &inf->dbinf;
     size_t		datasiz;
     
     if (S_ISREG(dbinf->stat.st_mode)) {
-      do_checksum(dbinf, path);
+      /* BUGGISH: don't test inf->did_sum, since this function's called
+       *          before do_check (very minorly buggish :)
+       */
+      do_checksum(opts, inf);
       inf->did_sum	 = 1;
       datasiz		 = sizeof(*dbinf);
     } else {
       datasiz		 = sizeof(dbinf->stat);
     }
-    if (cdb_put(db, path, pathlen, dbinf, datasiz) == -1)
+    if (cdb_put(db, inf->path, inf->pathlen, dbinf, datasiz) == -1)
       DIE("adding record to current-state db");
 }
 
 wft_ret_t process_file(const char *path, const struct stat *sb, void *data)
 {
     options	*opts		 = (options *) data;
-    fileinfo	inf;		/* file info for the DB */
+    fileinfo	inf;		/* initialize before use */
     checkset	cset;
     size_t	plen		 = strlen(path);
     char	*pathcopy	 = xstrdup(path);
     char	*p		 = pathcopy;
+#if defined(DEBUG) && 0
+    static int	counter;
+#endif
 
-    inf.did_sum	 = 0;
-    
     if (p[0] == '/' && p[1] == '/') {
       ++p;		/* pass the first of double initial slashes */
       --plen;
@@ -400,11 +436,15 @@ wft_ret_t process_file(const char *path, const struct stat *sb, void *data)
       --plen;
     }
 
-#if 0
-    printf("debug (%s): path (%s)\n", __FUNCTION__, p); /* debug */
+#if defined(DEBUG) && 0
+    fprintf(stderr, "debug (%s): count (%d) path (%s)\n",
+	    __FUNCTION__, ++counter, p); /* debug */
     usleep(5);			/* debug */
 #endif
 
+    /* initialize inf now that we've settled on a file path */
+    fileinfo_init(&inf, p, plen);
+
     cset	 = hashtbl_lookup(opts->ruleset, p, plen);
 
     /* see whether checkset includes boolean to ignore the file */
@@ -419,10 +459,10 @@ wft_ret_t process_file(const char *path, const struct stat *sb, void *data)
     memcpy(&inf.dbinf.stat, sb, sizeof(inf.dbinf.stat));
 
     if (opts->do_update)
-      do_update(opts, p, plen, &inf);
+      do_update(opts, &inf);
 
     if (opts->do_check)
-      do_check(opts, p, plen, &inf);
+      do_check(opts, &inf);
 
     free(pathcopy);
 
